from abc import ABC, abstractmethod
from typing import List, Any, Optional, Dict
import numpy as np
from ddi_fw.ml.model_wrapper import ModelWrapper, Result
from ddi_fw.ml.evaluation_helper import evaluate, Metrics
from ddi_fw.utils import utils
from ddi_fw import utils

import tensorflow as tf


def convert_to_categorical(arr, num_classes):
    """
    This function takes an array of labels and converts them to one-hot encoding 
    if they are not binary-encoded. If the array is already in a 
    compatible format, it returns the original array.

    Parameters:
    - arr: numpy array with label data (could be binary-encoded or label-encoded)
    - num_classes: number of classes to be used in one-hot encoding

    Returns:
    - The one-hot encoded array if the original array was binary or label encoded
    - The original array if it doesn't require any conversion
    """

    try:
        # First, check if the array is binary-encoded
        if not utils.is_binary_encoded(arr):
            # If the arr labels are binary-encoded, convert them to one-hot encoding
            return tf.keras.utils.to_categorical(np.argmax(arr, axis=1), num_classes=num_classes)
        else:
            print("No conversion needed, returning original array.")
            return arr
    except Exception as e:
        # If binary encoding check raises an error, print it and continue to label encoding check
        print(f"Error while checking binary encoding: {e}")

    try:
        # Check if the array is label-encoded
        if utils.is_label_encoded(arr):
            # If the arr labels are label-encoded, convert them to one-hot encoding
            return tf.keras.utils.to_categorical(arr, num_classes=num_classes)
    except Exception as e:
        # If label encoding check raises an error, print it
        print(f"Error while checking label encoding: {e}")
        # If the arr labels don't match any of the known encodings, raise an error
        raise ValueError("Unknown label encoding format.")

    # If no conversion was needed, return the original array

    return arr


class EnsembleStrategy(ABC):
    """Base class for ensemble strategies"""

    @abstractmethod
    def combine_predictions(self, predictions: List[np.ndarray]) -> np.ndarray:
        """Combine individual predictions into ensemble prediction"""
        pass

    @abstractmethod
    def get_strategy_name(self) -> str:
        """Return strategy name"""
        pass


class VotingStrategy(EnsembleStrategy):
    """Majority voting strategy for classification"""

    def combine_predictions(self, predictions: List[np.ndarray]) -> np.ndarray:
        import numpy as np
        from collections import Counter
        predictions_array = np.array(predictions)
        pred_labels = np.array([p.argmax(axis=1) for p in predictions_array])
        return np.apply_along_axis(
            lambda x: Counter(x).most_common(1)[0][0],
            axis=0,
            arr=pred_labels
        )

    def get_strategy_name(self) -> str:
        return "voting"


class AveragingStrategy(EnsembleStrategy):
    """Averaging strategy for regression or probability outputs"""

    def combine_predictions(self, predictions: List[np.ndarray]) -> np.ndarray:
        return np.mean(np.array(predictions), axis=0)

    def get_strategy_name(self) -> str:
        return "averaging"


class StackingStrategy(EnsembleStrategy):
    """Stacking strategy using a meta-learner"""

    def __init__(self, meta_learner: ModelWrapper, base_wrappers: List[ModelWrapper] = []):
        self.meta_learner = meta_learner
        self.base_wrappers = base_wrappers
        self.descriptor = "stacking_ensemble"

    def combine_predictions(self, predictions: List[np.ndarray]) -> np.ndarray:
        # Stack predictions as features for meta-learner
        self.fit()
        # self.meta_learner.fit(np.column_stack(predictions), None)
        # stacked_features = np.column_stack(predictions)
        return self.meta_learner.predict()

    def get_strategy_name(self) -> str:
        return "stacking"

    def fit(self):
        # Step 1: Train base models and collect softmax outputs
        inputs = []
        train_label, test_label = None, None
        train_idx_arr, val_idx_arr = None, None
        num_classes = None
        for wrapper in self.base_wrappers:
            train_label = wrapper.train_label
            test_label = wrapper.test_label

            # wrapper.train_data = self.train_data
            # wrapper.train_label = self.train_label
            train_idx_arr = wrapper.train_idx_arr
            val_idx_arr = wrapper.val_idx_arr
            # wrapper.train_idx_arr = self.train_idx_arr
            # wrapper.test_data = self.test_data
            # wrapper.test_label = self.test_label
            if wrapper.best_model is None:
                print(f"Training base model: {wrapper.descriptor}")
                best_model, _, val_preds_dict = wrapper.fit()
                num_classes = wrapper.num_classes
                wrapper.best_model = best_model
                # pred_train = wrapper.predict()
                # if val_preds_dict is not None:
                print("val_preds_dict")
                print(val_preds_dict)
                stacked_values = np.concatenate(list(val_preds_dict.values()))
            else:
                print(f"Using pre-trained base model: {wrapper.descriptor}")
                num_classes = wrapper.num_classes
                preds = [b.predict() for b in self.base_wrappers]
                stacked_values = np.stack(preds, axis=1)
                # stacked_values = np.hstack(
                #     [b.predict() for b in self.base_wrappers])
            # Use validation predictions for training meta-model
            inputs.append(stacked_values)

            # pred_val = wrapper.predict()
            # base_test_outputs.append(pred_val)

        # Step 2: Prepare meta-model input
        # shape: (n_samples, num_classes * num_base_models)
        # X_meta_train = np.hstack(inputs)
        X_meta_train = np.stack(inputs, axis=1)
        y_meta_train = train_label

        # Step 3: Train meta-model
        self.meta_learner.train_data = X_meta_train
        self.meta_learner.train_label = y_meta_train
        self.meta_learner.train_idx_arr = train_idx_arr
        self.meta_learner.val_idx_arr = val_idx_arr
        self.num_classes = num_classes
        # self.meta_learner.val_idx_arr = None
        preds = [b.predict() for b in self.base_wrappers]
        self.meta_learner.test_data = np.stack(preds, axis=1)

        # self.meta_learner.test_data = np.hstack(
        #     [b.predict() for b in self.base_wrappers])

        self.meta_learner.test_label = test_label

        print("Training meta-learner...")
        best_meta_model, best_meta_key, _ = self.meta_learner.fit()
        self.meta_learner.best_model = best_meta_model
        # self.best_model = best_meta_model

        return best_meta_model, best_meta_key

    def predict(self):
        # Step 3: Feed into meta-model
        pred = self.meta_learner.predict()
        return pred

    def fit_and_evaluate(self, print_detail=False):
        best_model, best_key = self.fit()
        pred = self.predict()
        pred_as_cat = convert_to_categorical(pred, self.num_classes)
        logs, metrics = evaluate(
            self.meta_learner.test_label, pred_as_cat, info=self.descriptor, print_detail=print_detail)
        metrics.format_float()
        return logs, metrics, pred


class GenericEnsembleWrapper(ModelWrapper):
    """Generic ensemble wrapper supporting multiple strategies"""

    STRATEGY_MAP = {
        "voting": VotingStrategy,
        "averaging": AveragingStrategy,
        "stacking": StackingStrategy
    }

    def __init__(
        self,
        base_wrappers: List[ModelWrapper],
        ensemble_strategies: str,          # now comma separated input
        name: str = "ensemble",
        meta_learner: Optional[ModelWrapper] = None,
        tracking_service: Optional[Any] = None,
        date: Optional[str] = None
    ):
        self.base_wrappers = base_wrappers
        self.name = name
        self.tracking_service = tracking_service
        self.date = date or utils.utc_time_as_string_simple_format()

        # Parse comma-separated strategies
        self.selected_strategies: List[EnsembleStrategy] = []
        for s in ensemble_strategies.split(","):
            s = s.strip().lower()
            if s not in self.STRATEGY_MAP:
                raise ValueError(f"Unknown strategy '{s}'")
            if s == "stacking":
                self.selected_strategies.append(StackingStrategy(meta_learner= meta_learner, base_wrappers= base_wrappers))
            else:
                self.selected_strategies.append(self.STRATEGY_MAP[s]())

        # prediction cache so we don't recompute
        self.wrapper_predictions: Dict[str, np.ndarray] = {}

        self.individual_metrics = {}
        self.ensemble_metrics: Dict[str, Metrics] = {}
        self.result = Result()

    # ---------------------------------------------------------
    # FIT
    # ---------------------------------------------------------
    def fit(self):
        train_label, test_label = None, None

        # if any(getattr(wrapper, 'best_model', None) is not None for wrapper in self.base_wrappers):
        #     # At least one wrapper has best_model set
        #     pass

        for wrapper in self.base_wrappers:
            train_label = wrapper.train_label
            test_label = wrapper.test_label

            best_model, _, _ = wrapper.fit()
            wrapper.best_model = best_model
            num_classes = wrapper.num_classes

        self.num_classes = num_classes
        self.test_label = test_label

    # ---------------------------------------------------------
    # PREDICT — with prediction caching
    # ---------------------------------------------------------
    def _compute_and_cache_predictions(self):
        """Compute wrapper predictions only once."""
        if self.wrapper_predictions:
            return  # already computed
        
        self.fit()

        for wrapper in self.base_wrappers:
            descriptor = wrapper.descriptor

            y_pred = wrapper.predict()

            # Store raw predictions (strategies want raw data)
            self.wrapper_predictions[descriptor] = y_pred

            # Evaluation for individual wrapper
            y_pred_cat = convert_to_categorical(y_pred, self.num_classes)
            _, wrapper_metrics = evaluate(self.test_label, y_pred_cat)
            wrapper_metrics.format_float()

            self.result.add_metric(descriptor, wrapper_metrics)
            self.individual_metrics[descriptor] = wrapper_metrics

    # ---------------------------------------------------------
    # MAIN PREDICTION PIPELINE
    # ---------------------------------------------------------
    def predict(self):
        # Compute predictions only once
        self._compute_and_cache_predictions()

        pred_list = list(self.wrapper_predictions.values())

        ensemble_outputs = {}

        for strategy in self.selected_strategies:
            strategy_name = strategy.get_strategy_name()

            # Combine predictions
            ensemble_pred = strategy.combine_predictions(pred_list)
            ensemble_pred = convert_to_categorical(
                ensemble_pred, self.num_classes)

            # Evaluate ensemble
            _, metrics = evaluate(self.test_label, ensemble_pred)
            metrics.format_float()

            token = f"ensemble_{strategy_name}"
            self.ensemble_metrics[token] = metrics
            self.result.add_metric(token, metrics)

            ensemble_outputs[token] = ensemble_pred

        return ensemble_outputs   # multiple ensemble predictions

    # ---------------------------------------------------------
    # FIT AND EVALUATE
    # ---------------------------------------------------------
    def fit_and_evaluate(self):
        self.fit()
        ensemble_preds = self.predict()
        return None, self.result, ensemble_preds

    # ---------------------------------------------------------
    def get_metrics_summary(self) -> Dict:
        return {
            "individual": self.individual_metrics,
            "ensemble": self.ensemble_metrics,
            "strategies": [s.get_strategy_name() for s in self.selected_strategies]
        }



from typing import List
import nltk
from nltk import sent_tokenize
from langchain_text_splitters.base import TextSplitter

nltk.download('punkt')

''' A text splitter that splits text into sentences using NLTK's sentence tokenizer.'''
class SentenceSplitter(TextSplitter):
    def split_text(self, text: str) -> List[str]:
        return sent_tokenize(text)


''' A text splitter that does not split the text at all, returning the entire text as a single chunk.'''
class PassthroughTextSplitter(TextSplitter):
    def split_text(self, text: str) -> List[str]:
        return [text]


# read from json
# programatic pipeline

import unittest
import json
from pathlib import Path
from ddi_fw.datasets.dataset_splitter import DatasetSplitter
from ddi_fw.pipeline import Pipeline, MultiPipeline
from ddi_fw.datasets import BaseDataset
from ddi_fw.langchain.embeddings import MeanPoolingStrategy, PoolingStrategy
from config.iris_dataset import IrisDataset
from tests.config.mnist_dataset import MNISTDataset

class TestPipeline(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Read the JSON configuration file
       
        
        config_path = Path('./tests/config/pipeline_test.json')
        with open(config_path, 'r') as f:
            cls.config = json.load(f)
            # print(cls.config)

    def test_pipeline_creation(self):
        dataset_splitter = DatasetSplitter()
        pooling_strategy = MeanPoolingStrategy()
        # dataset = IrisDataset(dataset_name="IrisDataset",
        #                       dataset_splitter_type = DatasetSplitter ,
        #                       index_path="tests/datasets/iris/indexes"
        #                    )
        # dataset.load()
        # dataset.split_dataset(save_indexes=True)
        
        cfg = self.config['experiments'][0]
        # Create a Pipeline object using the configuration
        mp = MultiPipeline(experiments_config=self.config).build()
        mp.run()
        print(f'name\t accuracy\tf-score\tprecision\trecall')
        for key in mp.results().keys():
            for m in mp.results()[key].metric_dict:
                accuracy = mp.results()[key].metric_dict[m].accuracy
                macro_f_score = mp.results()[key].metric_dict[m].f1_score['macro']
                macro_precision = mp.results()[key].metric_dict[m].precision['macro']
                macro_recall = mp.results()[key].metric_dict[m].recall['macro']
                print(f'{m}: {accuracy}\t{macro_f_score}\t{macro_precision}\t{macro_recall}')
            
    
    def test_pipeline_creation_mnist(self):
            
            x_config_path = Path('./tests/config/pipeline_test_mnist.json')
            with open(x_config_path, 'r') as f:
                config = json.load(f)
                print(config)
            
            dataset_splitter = DatasetSplitter()
            pooling_strategy = MeanPoolingStrategy()
            # dataset = MNISTDataset(dataset_name="MNISTDataset",
            #                     dataset_splitter_type = DatasetSplitter ,
            #                     index_path="tests/datasets/mnist/indexes"
            #                 )
            # dataset.load()
            # dataset.split_dataset(save_indexes=True)
            
            # cfg = self.config['experiments'][0]
            # # Create a Pipeline object using the configuration
            mp = MultiPipeline(experiments_config=config).build()
            mp.run()
            print(f'name\t accuracy\tf-score\tprecision\trecall')
            for key in mp.results().keys():
                for m in mp.results()[key].metric_dict:
                    accuracy = mp.results()[key].metric_dict[m].accuracy
                    macro_f_score = mp.results()[key].metric_dict[m].f1_score['macro']
                    macro_precision = mp.results()[key].metric_dict[m].precision['macro']
                    macro_recall = mp.results()[key].metric_dict[m].recall['macro']
                    print(f'{m}: {accuracy}\t{macro_f_score}\t{macro_precision}\t{macro_recall}')

 
    def test_pipeline_creation_openfda(self):
        import os
        from dotenv import load_dotenv
        import tensorflow as tf

        print("Num GPUs Available:", len(tf.config.list_physical_devices('GPU')))

        # Load variables from .env file into environment
        load_dotenv()

        # Set Kaggle credentials from environment
        os.environ['KAGGLE_USERNAME'] = os.getenv('KAGGLE_USERNAME')
        os.environ['KAGGLE_KEY'] = os.getenv('KAGGLE_KEY')

        # !kaggle datasets download -d kivancbayraktar/c8c6ab09-dbbd-4a57-8241-c81209bcbfcd --unzip -p /content/new_work/embeddings/all-embeddings
        # import kagglehub

        # # Download latest version
        # path = kagglehub.dataset_download("kivancbayraktar/c8c6ab09-dbbd-4a57-8241-c81209bcbfcd")

        # print("Path to dataset files:", path)
        # from ddi_fw.langchain import ChromaVectorStoreManager,BaseVectorStoreManager, FaissVectorStoreManager

        
        # faiss_vector_store_manager  = FaissVectorStoreManager(persist_directory=f"{path}/faiss/openfda-deneme-all-mpnet-base-v2/no_op")
        # # faiss_vector_store_manager  = FaissVectorStoreManager(persist_directory=f"{path}/embeddings/all-embeddings/faiss/openfda-deneme-all-mpnet-base-v2/no_op")

        # d = faiss_vector_store_manager.initialize_embedding_dict()
        # faiss_vector_store_manager.vector_store.index.ntotal
        x_config_path = Path('./tests/config/pipeline_test_openfda_2.json')
        with open(x_config_path, 'r') as f:
            config = json.load(f)
            # print(config)
        
        dataset_splitter = DatasetSplitter()
        pooling_strategy = MeanPoolingStrategy()
        # dataset = MNISTDataset(dataset_name="MNISTDataset",
        #                     dataset_splitter_type = DatasetSplitter ,
        #                     index_path="tests/datasets/mnist/indexes"
        #                 )
        # dataset.load()
        # dataset.split_dataset(save_indexes=True)
        
        # cfg = self.config['experiments'][0]
        # # Create a Pipeline object using the configuration
        mp = MultiPipeline(experiments_config=config).build()
        mp.run()
        print(f'name\t accuracy\tf-score\tprecision\trecall')
        for key in mp.results().keys():
            for m in mp.results()[key].metric_dict:
                accuracy = mp.results()[key].metric_dict[m].accuracy
                macro_f_score = mp.results()[key].metric_dict[m].f1_score['macro']
                macro_precision = mp.results()[key].metric_dict[m].precision['macro']
                macro_recall = mp.results()[key].metric_dict[m].recall['macro']
                print(f'{m}: {accuracy}\t{macro_f_score}\t{macro_precision}\t{macro_recall}')
                            
 
if __name__ == '__main__':
    unittest.main()


# from ner.ner import CTakesNER
# from datasets import CustomDataset

# from datasets.db_utils import create_connection

# import pandas as pd
# import numpy as np
# from datasets import IDF


# def __select_all_drugs_as_dataframe__(conn):

#     cur = conn.cursor()
#     cur.execute(f'''
#             select
#             _Drugs."index",
#             drugbank_id,
#             _Drugs.name,
#             description,
#             synthesis_reference,
#             indication,
#             pharmacodynamics,
#             mechanism_of_action,
#             toxicity,
#             metabolism,
#             absorption,
#             half_life,
#             protein_binding,
#             route_of_elimination,
#             volume_of_distribution,
#             clearance,
#             smiles,
#             smiles_morgan_fingerprint,
#             enzymes_polypeptides,
#             targets_polypeptides,
#             pathways,
#             tuis_description,
#             cuis_description,
#             entities_description

#             from _Drugs


#     ''')
#     #  where
#     #         targets_polypeptides is not null and
#     #         enzymes_polypeptides is not null and
#     #         pathways is not null and
#     #         smiles_morgan_fingerprint is not null
#     #        drugbank_id in {format(param)} and

#     # pathway is absent

#     rows = cur.fetchall()
#     headers = ['index', 'drugbank_id', 'name', 'description', 'synthesis_reference', 'indication', 'pharmacodynamics', 'mechanism_of_action', 'toxicity', 'metabolism', 'absorption', 'half_life',
#                'protein_binding', 'route_of_elimination', 'volume_of_distribution', 'clearance', 'smiles_notation', 'smile', 'enzyme', 'target', 'pathway',
#                'tui_description', 'cui_description', 'entities_description']
#     df = pd.DataFrame(columns=headers, data=rows)
#     df['smile'] = df['smile'].apply(lambda x:
#                                     np.fromstring(
#                                         x.replace(
#                                             '\n', '')
#                                         .replace('[', '')
#                                         .replace(']', '')
#                                         .replace('  ', ' '), sep=','))
#     df['enzyme'] = df['enzyme'].apply(
#         lambda x: x.split('|') if x is not None else [])
#     df['target'] = df['target'].apply(
#         lambda x: x.split('|') if x is not None else [])
#     df['pathway'] = df['pathway'].apply(
#         lambda x: x.split('|') if x is not None else [])
#     df['tui_description'] = df['tui_description'].apply(
#         lambda x: x.split('|') if x is not None else [])
#     df['cui_description'] = df['cui_description'].apply(
#         lambda x: x.split('|') if x is not None else [])
#     df['entities_description'] = df['entities_description'].apply(
#         lambda x: x.split('|') if x is not None else [])
#     return df


# db = f'datasets/custom/drugbank.db'
# conn = create_connection(db)
# drugs_df = __select_all_drugs_as_dataframe__(conn)
# index_path = 'datasets/custom/indexes'


# # ctakes_api_url = 'http://localhost:8080/ctakes-web-rest/service/analyze?pipeline=Default'
# # ctakes_api_url = 'http://localhost:8080/process?format=umls'
# # ctakes_api_url = 'http://10.10.10.29:8181/process?format=umls'
# ctakes_api_url = 'http://10.10.10.29:8181/analyze?pipeline=Default'
# ctakes_ner = CTakesNER(
#     drugs_df=drugs_df,
#     api_url=ctakes_api_url,
#     output_path='ner/output/ctakes',
#     columns=['description', 'indication'
#              , 'synthesis_reference', 'pharmacodynamics',
#              'mechanism_of_action', 'toxicity', 'metabolism', 'absorption', 'half_life', 'protein_binding', 'route_of_elimination','volume_of_distribution', 'clearance'],
#     ids=[])
# # ctakes_ner.run(run_for=['route_of_elimination'])
# # override columns
# # ctakes_ner.run(run_for=['description', 'indication', 'synthesis_reference', 'pharmacodynamics',
# #              'mechanism_of_action', 'toxicity', 'metabolism', 'absorption', 'half_life', 'protein_binding', 'volume_of_distribution', 'clearance'])
# # ner_df = ctakes_ner.create_dataframe(override=True)
# # print(ner_df.keys())
# # print(ner_df.shape)
# # print(ner_df.head())
# # print(ner_df.head())

# loaded_ner_df = ctakes_ner.load('ner/output/ctakes/ctakes_ner.pkl')
# print(loaded_ner_df.tail)
# print(loaded_ner_df.shape)
# print(loaded_ner_df[loaded_ner_df['drugbank_id'] == 'DB17386']['entities_clearance'])


from .ml_helper import MultiModalRunner
from .model_wrapper import ModelWrapper,Result
from .tensorflow_wrapper import TFModelWrapper
from .pytorch_wrapper import PTModelWrapper
from .ensemble_stacking_wrapper import StackingWrapper
from .ensemble_voting_wrapper import VotingWrapper
from .evaluation_helper import evaluate
from .tracking_service import TrackingService
from .ensemble_strategy import EnsembleStrategy, VotingStrategy, AveragingStrategy, StackingStrategy, GenericEnsembleWrapper


import os
import numpy as np
import pandas as pd
from scipy.spatial.distance import pdist, squareform
from sklearn.preprocessing import MultiLabelBinarizer

# todo pd.unique kullan
def find_distinct_elements(frame):
    # y = set(pd.unique(frame))
    y = set()
    for x in frame:
        if x is not None:
            for k in x:
                #     if type(k) == list:
                #         for i in k:
                #             y.add(i)
                #     else:
                y.add(k)
    return y


def find_distinct_elements_count(frame):
    y = set()
    for x in frame:
        if x is not None:
            y.update(x)
    return len(y)


class SimilarityMatrixGenerator:
    def __init__(self):
        # Check if GPU usage is enabled via an environment variable
        self.use_gpu = os.getenv("SIMILARITY_MATRIX_USE_GPU", "false").lower() == "true"

    def create_jaccard_similarity_matrices_ex_1(self, array):
        jaccard_sim = 1 - pdist(array, metric='jaccard')
        jaccard_sim_matrix = squareform(jaccard_sim)
        return jaccard_sim_matrix

    # https://github.com/YifanDengWHU/DDIMDL/blob/master/DDIMDL.py , def Jaccard(matrix):
    def create_jaccard_similarity_matrices_ex_2(self, matrix)->np.ndarray:
        matrix = np.asmatrix(matrix)
        numerator = matrix * matrix.T
        denominator = np.ones(np.shape(matrix)) * matrix.T + \
            matrix * np.ones(np.shape(matrix.T)) - matrix * matrix.T
        matrix = numerator / denominator
        return np.nan_to_num(matrix, nan=0.0)
        # return matrix
        
        
    def create_jaccard_similarity_matrices(self, matrix: np.ndarray)->np.ndarray:
        if self.use_gpu:
            print("Using GPU for Jaccard similarity matrix computation.")
            return self.__create_jaccard_similarity_matrices_gpu(matrix)
        else:
            return self.__create_jaccard_similarity_matrices(matrix)
        
    """produced from ChatGPT"""
    def __create_jaccard_similarity_matrices(self, matrix: np.ndarray)->np.ndarray:
        """
        Efficiently compute the Jaccard similarity between rows of a binary matrix using vectorized operations.

        Parameters:
        matrix (np.ndarray): A 2D binary NumPy array (only 0s and 1s).

        Returns:
        np.ndarray: A 2D NumPy array containing the pairwise Jaccard similarity.
        """
        if not ((matrix == 0) | (matrix == 1)).all():
            raise ValueError("Input matrix must be binary (contain only 0s and 1s).")

        # Intersection: dot product (each pair of rows)
        intersection = matrix @ matrix.T

        # Row-wise sum (number of 1s per row)
        row_sums = matrix.sum(axis=1, keepdims=True)

        # Union: |A ∪ B| = |A| + |B| - |A ∩ B|
        union = row_sums + row_sums.T - intersection

        # Avoid division by zero
        with np.errstate(divide='ignore', invalid='ignore'):
            similarity = intersection / union
            similarity[np.isnan(similarity)] = 1.0  # If both rows are all zeros, define similarity as 1

        return similarity


    def __create_jaccard_similarity_matrices_gpu(self,matrix: np.ndarray) -> np.ndarray:
        """
        Efficiently compute the Jaccard similarity between rows of a binary matrix using GPU-accelerated CuPy.

        Parameters:
        matrix (cp.ndarray): A 2D binary CuPy array (only 0s and 1s).

        Returns:
        cp.ndarray: A 2D CuPy array containing the pairwise Jaccard similarity.
        """
        try:
            import cupy as cp
        except ImportError:
            raise ImportError("cupy is required for GPU Jaccard similarity computation. Please install it with 'pip install cupy'.")
                
                
        if not ((matrix == 0) | (matrix == 1)).all():
            raise ValueError("Input matrix must be binary (contain only 0s and 1s).")
        matrix = cp.asarray(matrix)
        # Intersection: dot product (each pair of rows)
        intersection = matrix @ matrix.T

        # Row-wise sum (number of 1s per row)
        row_sums = matrix.sum(axis=1, keepdims=True)

        # Union: |A ∪ B| = |A| + |B| - |A ∩ B|
        union = row_sums + row_sums.T - intersection

        # Avoid division by zero
        similarity = cp.divide(intersection, union, out=cp.ones_like(intersection, dtype=cp.float64)) #, where=union != 0

        return cp.asnumpy(similarity)



class VectorGenerator:
    def __init__(self, df):
        self.df = df

    # https://github.com/YifanDengWHU/DDIMDL/blob/master/DDIMDL.py#L86
    # def generate_feature_vector(self, column):
    #     # Initialize list to store all distinct features across all rows
    #     all_features = []
        
    #     # Loop through the column to extract features, split by '|', and collect all distinct ones
    #     drug_list = np.array(self.df[column]).tolist()
    #     for i in drug_list:
    #         for each_feature in i.split('|'):
    #             if each_feature not in all_features:
    #                 all_features.append(each_feature)
        
    #     # Initialize a matrix to hold feature vectors (rows for each element, columns for each distinct feature)
    #     feature_matrix = np.zeros((len(drug_list), len(all_features)), dtype=float)
        
    #     # Create a DataFrame to store the feature matrix with the column names as the distinct features
    #     df_feature = pd.DataFrame(feature_matrix, columns=all_features)
        
    #     # Fill the feature matrix (set value to 1 if feature is present for the specific item in the column)
    #     for i in range(len(drug_list)):
    #         for each_feature in drug_list[i].split('|'):
    #             if each_feature in all_features:
    #                 df_feature[each_feature].iloc[i] = 1
        
    #     # Convert DataFrame to numpy array and return
    #     print("Feature vectors generated")
    #     return df_feature.to_numpy()

    def generate_feature_vector(self, column):
        bit_vectors = []
        map = dict()
        idx = 0
        count = find_distinct_elements_count(self.df[column])
        print(f"{column} has {count} different items")
        for ind in self.df.index:
            e = self.df[column][ind]
            # vector = np.zeros(len(sorted_features))
            vector = np.zeros(count)
            if e is not None:
                for item in e:
                    if item in map:
                        vector[map[item]] = 1
                    else:
                        vector[idx]=1
                        map[item] = idx
                        idx += 1 
 
            bit_vectors.append(vector)
        print("array oluşturuldu")
        return np.array(bit_vectors)
     
    
    # def generate_feature_vector(self, column):
    #     bit_vectors = []
    #     distinct_feature = find_distinct_elements(self.df[column])
    #     sorted_features = sorted(distinct_feature)
    #     for ind in self.df.index:
    #         e = self.df[column][ind]
    #         vector = np.zeros(len(sorted_features))
    #         if e is not None:
    #             indexes = [i for i, x in enumerate(sorted_features) if x in e]
    #             np.put(vector, indexes, np.ones(len(indexes)))
    #         bit_vectors.append(vector)
    #     return bit_vectors

# bit_vectors ndarray olacak
    def generate_feature_vectors(self, columns):
        vectors = dict()
        for column in columns:
            bit_vectors = self.generate_feature_vector(column)
            vectors[column] = bit_vectors
        return vectors
    


# generate feature vector
# np.hstack

# https://www.datasciencelearner.com/how-to-create-an-array-of-bits-in-python/
#
